/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.itvsh.kop.common.binaryfile;

import java.io.IOException;
import java.io.InputStream;
import java.util.Objects;
import java.util.Optional;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicBoolean;
import java.util.function.Function;

import org.apache.commons.io.IOUtils;

import de.itvsh.kop.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;
import lombok.AccessLevel;
import lombok.NoArgsConstructor;
import lombok.NonNull;
import lombok.extern.log4j.Log4j2;

@Log4j2
@NoArgsConstructor(access = AccessLevel.PRIVATE)
public class GrpcFileUploadUtils {

	static final int CHUNK_SIZE = 255 * 1024;

	/*
	 * Q = Request Type; S = Response Type
	 */
	public static <Q, S> FileSender<Q, S> createSender(Function<byte[], Q> chunkBuilder, InputStream inputStream,
			Function<StreamObserver<S>, CallStreamObserver<Q>> reqObserverBuilder) {
		return new FileSender<>(chunkBuilder, reqObserverBuilder, inputStream);
	}

	public static class FileSender<Q, S> {
		private final Function<byte[], Q> chunkBuilder;
		private final InputStream inputStream;

		private final CompletableFuture<S> resultFuture = new CompletableFuture<>();
		private final Function<StreamObserver<S>, CallStreamObserver<Q>> reqObserverBuilder;
		private CallStreamObserver<Q> requestObserver;

		private Optional<Q> metaData = Optional.empty();
		private final AtomicBoolean metaDataSent = new AtomicBoolean(false);
		private final AtomicBoolean done = new AtomicBoolean(false);

		FileSender(Function<byte[], Q> chunkBuilder, Function<StreamObserver<S>, CallStreamObserver<Q>> reqObserverBuilder, InputStream inputStream) {
			this.chunkBuilder = chunkBuilder;
			this.inputStream = IOUtils.buffer(inputStream, CHUNK_SIZE);
			this.reqObserverBuilder = reqObserverBuilder;
		}

		public FileSender<Q, S> withMetaData(@NonNull Q metaData) {
			this.metaData = Optional.of(metaData);
			return this;
		}

		public CompletableFuture<S> send() {
			LOG.debug("Start sending File.");
			var responseObserver = BinaryFileUploadStreamObserver.create(resultFuture, this::sendNext);
			requestObserver = reqObserverBuilder.apply(responseObserver);

			return resultFuture;
		}

		void sendNext() {
			if (!done.get()) {
				waitForOberver();
				LOG.debug("Sending next chunk.");
				sendMetaData();
				long sentSize = sendNextChunk();
				checkForEndOfStream(sentSize);
			}
		}

		private void waitForOberver() {
			synchronized (this) {
				while (Objects.isNull(requestObserver)) {
					try {
						LOG.debug("wait for observer");
						wait(300);
					} catch (InterruptedException e) {
						LOG.error("Error on waiting for request Observer.", e);
						Thread.currentThread().interrupt();
					}
				}
			}

		}

		long sendNextChunk() {
			byte[] contentToSend = readFromStream();

			if (contentToSend.length > 0) {
				sendChunk(contentToSend);
			}
			return contentToSend.length;
		}

		void sendChunk(byte[] content) {
			LOG.debug("Sending {} byte Data.", content.length);
			var chunk = chunkBuilder.apply(content);
			requestObserver.onNext(chunk);
		}

		byte[] readFromStream() {
			try {
				return inputStream.readNBytes(CHUNK_SIZE);
			} catch (IOException e) {
				throw new TechnicalException("Error on sending a single chunk", e);
			}
		}

		void sendMetaData() {
			metaData.filter(md -> !metaDataSent.get()).ifPresent(this::doSendMetaData);
		}

		private void doSendMetaData(Q metadata) {
			LOG.debug("Sending Metadata.");
			requestObserver.onNext(metadata);
			metaDataSent.set(true);
		}

		void checkForEndOfStream(long sentSize) {
			if (sentSize < CHUNK_SIZE) {
				LOG.debug("File Transfer done. Closing stream.");
				IOUtils.closeQuietly(inputStream);
				requestObserver.onCompleted();
				done.set(true);
				IOUtils.closeQuietly(inputStream);
			}
		}
	}

}