/*
 * Copyright (C) 2022 Das Land Schleswig-Holstein vertreten durch den
 * Ministerpräsidenten des Landes Schleswig-Holstein
 * Staatskanzlei
 * Abteilung Digitalisierung und zentrales IT-Management der Landesregierung
 *
 * Lizenziert unter der EUPL, Version 1.2 oder - sobald
 * diese von der Europäischen Kommission genehmigt wurden -
 * Folgeversionen der EUPL ("Lizenz");
 * Sie dürfen dieses Werk ausschließlich gemäß
 * dieser Lizenz nutzen.
 * Eine Kopie der Lizenz finden Sie hier:
 *
 * https://joinup.ec.europa.eu/collection/eupl/eupl-text-eupl-12
 *
 * Sofern nicht durch anwendbare Rechtsvorschriften
 * gefordert oder in schriftlicher Form vereinbart, wird
 * die unter der Lizenz verbreitete Software "so wie sie
 * ist", OHNE JEGLICHE GEWÄHRLEISTUNG ODER BEDINGUNGEN -
 * ausdrücklich oder stillschweigend - verbreitet.
 * Die sprachspezifischen Genehmigungen und Beschränkungen
 * unter der Lizenz sind dem Lizenztext zu entnehmen.
 */
package de.itvsh.kop.common.binaryfile;

import static org.assertj.core.api.Assertions.*;
import static org.mockito.ArgumentMatchers.*;
import static org.mockito.Mockito.*;

import java.io.IOException;
import java.io.InputStream;
import java.util.concurrent.CompletableFuture;
import java.util.function.Function;

import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Nested;
import org.junit.jupiter.api.Test;
import org.mockito.ArgumentCaptor;
import org.mockito.Captor;
import org.mockito.Mock;
import org.mockito.Spy;

import de.itvsh.kop.common.binaryfile.BinaryFileTestFactory.TestRequestType;
import de.itvsh.kop.common.binaryfile.BinaryFileTestFactory.TestResponseType;
import de.itvsh.kop.common.binaryfile.GrpcFileUploadUtils.FileSender;
import de.itvsh.kop.common.errorhandling.TechnicalException;
import io.grpc.stub.CallStreamObserver;
import io.grpc.stub.StreamObserver;

class GrpcFileUploadUtilsTest {

	@Spy
	private GrpcFileUploadUtils service;

	@Mock
	private Function<byte[], TestRequestType> chunkBuilder;
	@Mock
	private Function<StreamObserver<TestResponseType>, CallStreamObserver<TestRequestType>> reqObserverBuilder;
	@Mock
	private CallStreamObserver<TestRequestType> requestObserver;
	@Mock
	private InputStream inputStream;

	private FileSender<TestRequestType, TestResponseType> fileSender;

	@Mock
	private TestRequestType metaData;

	@BeforeEach
	void init() {
		when(reqObserverBuilder.apply(any())).thenReturn(requestObserver);
		fileSender = spy(GrpcFileUploadUtils.createSender(chunkBuilder, inputStream, reqObserverBuilder));
	}

	@Nested
	class TestCreateFileSender {

		@Test
		void shouldCreateRequestObserver() {
			GrpcFileUploadUtils.createSender(chunkBuilder, inputStream, reqObserverBuilder);

			verify(reqObserverBuilder, atLeastOnce()).apply(notNull());
		}
	}

	@Nested
	class TestSendBinaryFile {

		@Captor
		private ArgumentCaptor<Runnable> runnableCaptor;

		@Test
		void shouldSetOnReadyHandler() {
			callService();

			verify(requestObserver).setOnReadyHandler(any());
		}

		@Test
		void shouldCallSendNextOnReady() {
			callService();

			verify(requestObserver).setOnReadyHandler(runnableCaptor.capture());
			runnableCaptor.getValue().run();

			verify(fileSender, timeout(500)).sendNext();
		}

		@Test
		void shouldReturnFuture() {
			var result = callService();

			assertThat(result).isNotNull();
		}

		private CompletableFuture<TestResponseType> callService() {
			return fileSender.send();
		}
	}

	@Nested
	class TestSendNext {

		private static final long SENT_SIZE = 1024;

		@Test
		void shouldCallSendMetaData() {
			fileSender.sendNext();

			verify(fileSender).sendMetaData();
		}

		@Test
		void shouldSendNextChunk() {
			fileSender.sendNext();

			verify(fileSender).sendNextChunk();
		}

		@Test
		void shouldCheckForEndOfStream() {
			when(fileSender.sendNextChunk()).thenReturn(SENT_SIZE);
			
			fileSender.sendNext();

			verify(fileSender).checkForEndOfStream(SENT_SIZE);
		}

	}

	@Nested
	class TestSendNextChunk {

		private static final byte[] CHUNK_PART = "ChunkPartContent".getBytes();

		@BeforeEach
		void init() {
			doReturn(CHUNK_PART).when(fileSender).readFromStream();
		}

		@Test
		void shouldReadData() {
			callService();

			verify(fileSender).readFromStream();
		}

		@Test
		void shouldSendNextChunk() {
			callService();

			verify(fileSender).sendChunk(CHUNK_PART);
		}

		@Test
		void shouldReturnSize() {
			var result = callService();

			assertThat(result).isEqualTo(CHUNK_PART.length);
		}

		private long callService() {
			return fileSender.sendNextChunk();
		}
	}

	@Nested
	class TestSendChunk {

		private static final byte[] CHUNK_PART = "ChunkPartContent".getBytes();

		@Test
		void shouldApplyBuildChunk() throws IOException {
			callService();

			verify(chunkBuilder).apply(CHUNK_PART);
		}

		@Test
		void shouldCallOnNext() throws IOException {
			callService();

			verify(requestObserver).onNext(any());
		}

		private void callService() {
			fileSender.sendChunk(CHUNK_PART);
		}
	}

	@Nested
	class TestSendMetaData {

		@Test
		void shouldNotSendWithoutMetadata() {
			fileSender.sendMetaData();

			verify(requestObserver, never()).onNext(any());
		}

		@Test
		void shouldSendMetadata() {
			fileSender.withMetaData(metaData).sendMetaData();

			verify(requestObserver).onNext(metaData);
		}

		@Test
		void shouldSendMetadataOnlyOnce() {
			fileSender.withMetaData(metaData).sendMetaData();
			fileSender.sendMetaData();

			verify(requestObserver).onNext(metaData);
		}
	}

	@Nested
	class TestReadFromStream {
		@Test
		void shouldThrowException() throws IOException {
			doThrow(IOException.class).when(inputStream).read(any(), anyInt(), anyInt());

			assertThatThrownBy(() -> fileSender.readFromStream()).isInstanceOf(TechnicalException.class);
		}
	}

}